#include "stdafx.h"
#include "PrimeTables.h"
#include "Factorize.h"
#include <algorithm>
#include <fstream>
#include <process.h>
#include <unordered_map>

#undef max

#pragma pack(push, 1)
struct PrimeData
{
	PrimeData(unsigned int _p, number _inverse, number _max_value) : p(_p), inverse(_inverse), max_value(_max_value) {}

	unsigned int p;
	number inverse;
	number max_value;
};
#pragma pack(pop)

static struct FoundSums
{
	typedef number Element;

	enum Params : number
	{
		// If (a*u, a) is a breeder then (S(a)-a)/a = (u - 1) / S(u)
		// Set the limits for the search: a <= 10^13, u <= 10^15
		LimitA = 10000000000000,
		LimitU = LimitA * 100,

		// 64 GB of data is enough to store 2^33 64-bit numbers
		// If your PC has less than 64 GB of RAM, set SizeInGb to a more appropriate value!!!
		SizeInGb = 4,

		// Make sure that last steps of binary search fit into one 4 KB memory page
		BSearchLastStepSizeInBytes = 4096,

		// Look-up table to speed up the beginning of binary search
		// It reduces search range to at most 2^33 / 16777216 = 512 64-bit numbers which fit into one 4 KB memory page
		BSearchBeginTableSize = ((number(1) << 33) / BSearchLastStepSizeInBytes) * sizeof(Element),

		// 1024 buckets, 8388608 elements per bucket on average
		BucketBits = 10,
	};

	std::vector<PrimeData> primes;

	Element* begin;
	Element* end;
	number count;

	HANDLE hSumsFile;
	HANDLE hSumsMap;

	number bsearchBegin[BSearchBeginTableSize];

	static bool StopSearchURanges;

	struct Bucket
	{
		Element data[1048576];
		number dataCount;
		HANDLE hOutputFile;

		void Init(number index)
		{
			char name[64];
			sprintf_s(name, "bucket_%I64u.dat", index);
			hOutputFile = CreateFile(name, GENERIC_WRITE, FILE_SHARE_READ, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, 0);
			if (hOutputFile == INVALID_HANDLE_VALUE)
			{
				std::cerr << "CreateFile returned error " << GetLastError() << std::endl;
				__debugbreak();
			}
		}

		void DeInit()
		{
			Flush();
			CloseHandle(hOutputFile);
		}

		void Flush()
		{
			if (dataCount > 0)
			{
				DWORD size = static_cast<DWORD>(sizeof(Element) * dataCount);
				char* p = reinterpret_cast<char*>(data);
				while (size)
				{
					DWORD k;
					if (!WriteFile(hOutputFile, p, size, &k, 0))
					{
						std::cerr << "WriteFile returned error " << GetLastError() << std::endl;
						__debugbreak();
					}
					p += k;
					size -= k;
				}
			}
			dataCount = 0;
		}
	};

	Bucket* buckets;

	NOINLINE void InitBSearchBegin(const number index, const number a, const number b)
	{
		if (index < BSearchBeginTableSize)
		{
			const number c = (a + b) >> 1;
			bsearchBegin[index] = begin[c];
			InitBSearchBegin(index * 2 + 1, c + 1, b);
			InitBSearchBegin(index * 2 + 2, a, c);
		}
	}
	
	static BOOL WINAPI CtrlBreakHandler(_In_ DWORD dwCtrlType)
	{
		if ((dwCtrlType == CTRL_C_EVENT) || (dwCtrlType == CTRL_BREAK_EVENT))
		{
			StopSearchURanges = true;
			return true;
		}
		return false;
	}

	NOINLINE void Init(bool generateSums)
	{
		SetConsoleCtrlHandler(CtrlBreakHandler, true);
		primes.reserve(1951957);
		for (unsigned int p = 2; p < SearchLimit::SQRT2; ++p)
		{
			if (IsPrime(p))
			{
				number x1 = number(-1);
				number x2 = 1;
				if (p > 2)
				{
					number v1 = ~number(p) + 1;
					number v2 = p;
					while (v2 > 1)
					{
						const number q = v1 / v2;
						const number x3 = x1 - q * x2;
						const number v3 = v1 - q * v2;
						x1 = x2;
						x2 = x3;
						v1 = v2;
						v2 = v3;
					}
				}
				primes.push_back(PrimeData(p, x2, number(-1) / p));
			}
		}

		if (generateSums)
		{
			buckets = reinterpret_cast<Bucket*>(VirtualAlloc(0, sizeof(Bucket) << BucketBits, MEM_COMMIT, PAGE_READWRITE));
			for (number i = 0; i < (1 << BucketBits); ++i)
				buckets[i].Init(i);
			GenerateSumsA(1, 1, SearchLimit::SQRT2, 0);
			for (number i = 0; i < 1 << BucketBits; ++i)
				buckets[i].DeInit();
			VirtualFree(buckets, 0, MEM_RELEASE);
			puts("");

			hSumsFile = CreateFile("sums.dat", GENERIC_WRITE, FILE_SHARE_READ, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, 0);
			if (hSumsFile == INVALID_HANDLE_VALUE)
			{
				std::cerr << "CreateFile returned error " << GetLastError() << std::endl;
				__debugbreak();
			}
			std::vector<number> bucketData;
			for (number i = 0; i < (1 << BucketBits); ++i)
			{
				char bucketFileName[64];
				sprintf_s(bucketFileName, "bucket_%I64u.dat", i);
				HANDLE hBucketFile = CreateFile(bucketFileName, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, 0);
				if (hBucketFile == INVALID_HANDLE_VALUE)
				{
					std::cerr << "CreateFile returned error " << GetLastError() << std::endl;
					__debugbreak();
				}
				LARGE_INTEGER bucketFileSize;
				if (!GetFileSizeEx(hBucketFile, &bucketFileSize))
				{
					std::cerr << "GetFileSizeEx returned error " << GetLastError() << std::endl;
					__debugbreak();
				}
				if (bucketFileSize.QuadPart % sizeof(Element))
				{
					std::cerr << "Incorrect file size for " << bucketFileName << std::endl;
					__debugbreak();
				}
				const number numElements = bucketFileSize.QuadPart / sizeof(Element);
				printf("Bucket %I64u: %I64u elements          \r", i, numElements);
				bucketData.resize(numElements);
				number size = static_cast<number>(bucketFileSize.QuadPart);
				char* p = reinterpret_cast<char*>(bucketData.data());
				while (size)
				{
					DWORD n = 1048576;
					if (n > size)
						n = static_cast<DWORD>(size);
					DWORD k;
					if (!ReadFile(hBucketFile, p, n, &k, 0))
					{
						std::cerr << "ReadFile returned error " << GetLastError() << std::endl;
						__debugbreak();
					}
					p += k;
					size -= k;
				}
				CloseHandle(hBucketFile);
				DeleteFile(bucketFileName);
				std::sort(bucketData.begin(), bucketData.end());
				size = static_cast<number>(bucketFileSize.QuadPart);
				p = reinterpret_cast<char*>(bucketData.data());
				while (size)
				{
					DWORD k;
					DWORD n = 1048576;
					if (n > size)
						n = static_cast<DWORD>(size);
					if (!WriteFile(hSumsFile, p, n, &k, 0))
					{
						std::cerr << "WriteFile returned error " << GetLastError() << std::endl;
						__debugbreak();
					}
					p += k;
					size -= k;
				}
			}
			printf("\n");
			CloseHandle(hSumsFile);
		}

		hSumsFile = CreateFile("sums.dat", GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_RANDOM_ACCESS, 0);
		if (hSumsFile == INVALID_HANDLE_VALUE)
		{
			std::cerr << "CreateFile returned error " << GetLastError() << std::endl;
			__debugbreak();
		}
		LARGE_INTEGER sumsFileSize;
		if (!GetFileSizeEx(hSumsFile, &sumsFileSize))
		{
			std::cerr << "GetFileSizeEx returned error " << GetLastError() << std::endl;
			__debugbreak();
		}
		if (sumsFileSize.QuadPart % sizeof(Element))
		{
			std::cerr << "Incorrect file size for sums.dat" << std::endl;
			__debugbreak();
		}
		count = sumsFileSize.QuadPart / sizeof(Element);

		hSumsMap = CreateFileMapping(hSumsFile, 0, PAGE_READONLY, 0, 0, nullptr);
		if (!hSumsMap)
		{
			std::cerr << "CreateFileMapping returned error " << GetLastError() << std::endl;
			__debugbreak();
		}
		begin = reinterpret_cast<Element*>(MapViewOfFile(hSumsMap, FILE_MAP_READ, 0, 0, 0));
		if (!begin)
		{
			std::cerr << "MapViewOfFile returned error " << GetLastError() << std::endl;
			__debugbreak();
		}
		end = begin + count;

		// Read sums into memory
		volatile Element tmp;
		for (Element* el = begin; el < end; el += 4096 / sizeof(Element))
		{
			tmp = *el;
			const number index = static_cast<number>(el - begin);
			if ((index & 4194303) == 0)
				printf("Loading sums.dat: %.1f%%\r", index * 100.0 / (end - begin));
		}

		printf("Loading sums.dat: 100.0%%\nInitializing binary search data...");
		InitBSearchBegin(0, 0, count - 1);
		printf("done\nPress Ctrl+C to stop searching \"u\" values and start the next step\n");
	}

	NOINLINE void GenerateSumsA(const number aValue, const number aSumOfDivisors, const number aPreviousPrimeFactor, const number aRecursionLevel)
	{
		if (aRecursionLevel > 0)
		{
			// Store aSumOfDivisors / aValue as fixed point value
			// 1.5 < S(n)/n < 2
			// 0.5 < (S(n)-n)/n < 1
			const number a = aSumOfDivisors - aValue;
			if (a * 2 > aValue)
			{
				number r;
				// 0.5 < a / aValue < 1
				// 1 < (a * 2) / aValue < 2
				// 0 < (a * 2) / aValue - 1 < 1
				// 0 < (a * 2 - aValue) / aValue < 1
				number q = udiv128(a * 2 - aValue, 0, aValue, &r);
				if (r * 2 > aValue)
					++q;

				Bucket& curBucket = buckets[q >> (64 - BucketBits)];
				if (curBucket.dataCount == ARRAYSIZE(curBucket.data))
					curBucket.Flush();
				curBucket.data[curBucket.dataCount++] = q;

				++count;
				if ((count & 1048575) == 0)
					printf("%I64uM values generated          \r", count >> 20);

				if (count >= (FoundSums::SizeInGb << 30) / sizeof(FoundSums::Element))
				{
					printf("\nLast \"a\" in the table: %I64u", aValue);
					return;
				}
			}
		}

		if (aPreviousPrimeFactor == 3)
			return;

		for (const PrimeData& curPrime : primes)
		{
			if (curPrime.p >= aPreviousPrimeFactor)
				break;

			number highProductNextValue;
			number nextValue = _umul128(aValue, curPrime.p, &highProductNextValue);
			if ((nextValue > LimitA) || highProductNextValue)
				break;

			number n = aSumOfDivisors * curPrime.p;
			number sumN = aSumOfDivisors + n;
			for (;;)
			{
				if (sumN < nextValue * 2)
					GenerateSumsA(nextValue, sumN, curPrime.p, aRecursionLevel + 1);

				if (count >= (FoundSums::SizeInGb << 30) / sizeof(FoundSums::Element))
					return;

				nextValue = _umul128(nextValue, curPrime.p, &highProductNextValue);
				if ((nextValue > LimitA) || highProductNextValue)
					break;

				n *= curPrime.p;
				sumN += n;
			}
		}
	}

	NOINLINE void FindBreeders(const number aValue, const number aSumOfDivisors, const number aPreviousPrimeFactor, const number aRecursionLevel, std::ofstream& breeders, const std::unordered_map<number, std::vector<std::pair<number, number>>>& aFoundU)
	{
		if (aRecursionLevel > 0)
		{
			const number a = aSumOfDivisors - aValue;
			if (a * 2 > aValue)
			{
				number r;
				number q = udiv128(a * 2 - aValue, 0, aValue, &r);
				if (r * 2 > aValue)
					++q;

				auto it = aFoundU.find(q);
				if (it != aFoundU.end())
				{
					for (std::pair<number, number> u : it->second)
					{
						if (aValue == u.first)
							continue;

						const number au_mod_6 = (aValue % 6) * (u.first % 6);
						if ((au_mod_6 % 6) == 0)
							continue;

						// (S(a)-a)*S(u) = (u - 1)*a
						number n1[2];
						number n2[2];
						n1[0] = _umul128(aSumOfDivisors - aValue, u.second, &n1[1]);
						n2[0] = _umul128(u.first - 1, aValue, &n2[1]);
						if ((n1[0] != n2[0]) || (n1[1] != n2[1]))
							continue;

						if (GCD(aValue, u.first) != 1)
							continue;

						OutputBreeder(breeders, u.first, aValue);
					}
					breeders.flush();
				}
				++count;
				if ((count & 1048575) == 0)
					printf("%I64uM sums checked            \r", count >> 20);

				if (count >= (FoundSums::SizeInGb << 30) / sizeof(FoundSums::Element))
				{
					printf("\nLast \"a\" in the table: %I64u", aValue);
					return;
				}
			}
		}

		if (aPreviousPrimeFactor == 3)
			return;

		for (const PrimeData& curPrime : primes)
		{
			if (curPrime.p >= aPreviousPrimeFactor)
				break;

			number highProductNextValue;
			number nextValue = _umul128(aValue, curPrime.p, &highProductNextValue);
			if ((nextValue > LimitA) || highProductNextValue)
				break;

			number n = aSumOfDivisors * curPrime.p;
			number sumN = aSumOfDivisors + n;
			for (;;)
			{
				if (sumN < nextValue * 2)
					FindBreeders(nextValue, sumN, curPrime.p, aRecursionLevel + 1, breeders, aFoundU);

				if (count >= (FoundSums::SizeInGb << 30) / sizeof(FoundSums::Element))
					return;

				nextValue = _umul128(nextValue, curPrime.p, &highProductNextValue);
				if ((nextValue > LimitA) || highProductNextValue)
					break;

				n *= curPrime.p;
				sumN += n;
			}
		}
	}

	FORCEINLINE number FindSum(const number q) const
	{
		number a = 0;
		number b = count - 1;
		number index = 0;
		while (index < BSearchBeginTableSize)
		{
			const number c = (a + b) >> 1;
			if (bsearchBegin[index] < q)
			{
				a = c + 1;
				index = index * 2 + 1;
			}
			else
			{
				b = c;
				index = index * 2 + 2;
			}
		}

		while (a < b)
		{
			const number c = (a + b) >> 1;
			if (begin[c] < q)
				a = c + 1;
			if (begin[c] >= q)
				b = c;
		}

		return (begin[b] == q) ? b : number(-1);
	}

	FORCEINLINE static void IncrementMod6(number& n)
	{
		++n;
		if (n == 6)
			n = 0;
	}

	struct URangeData
	{
		static volatile unsigned long long counter;

		number threadIndex;
		number numThreads;

		FoundSums* pThis;
		number step;
		std::vector<std::pair<number, number>> sumsU;
		std::fstream* breeders;
	};

	NOINLINE static unsigned int WINAPI SearchURangeThreadProc(void* ptr)
	{
		URangeData* data = (URangeData*)ptr;
		while (!StopSearchURanges)
		{
			const number counterValue = _InterlockedIncrement(&URangeData::counter);
			const number u_begin = (counterValue - 1) * data->step + 1;
			const number u_end = counterValue * data->step;
			if (u_end >= LimitU)
				return 0;

			if ((u_end / 100000000) != (u_begin / 100000000))
				printf("\ru = %.1f*10^9", u_end / 1e9);
			data->pThis->SearchURange(u_begin, u_end, data->sumsU, *data->breeders);
		}
		data->breeders->close();
		return 0;
	}

	NOINLINE void SearchURange(const number u_begin, const number u_end, std::vector<std::pair<number, number>>& sumsU, std::fstream& breeders) const
	{
		sumsU.clear();
		sumsU.reserve(u_end - u_begin);
		for (number u = u_begin; u < u_end; ++u)
			sumsU.push_back(std::pair<number, number>(u, 1));

		for (const PrimeData& curPrime : primes)
		{
			if (curPrime.p == 2)
			{
				number u = u_begin & 1;
				if (u == 0)
					u = u_begin;
				else
					u = u_begin + 2 - 1;
				while (u < u_end)
				{
					unsigned long index;
					_BitScanForward64(&index, u);
					std::pair<number, number>& curSumU = sumsU[u - u_begin];
					curSumU.first >>= index;
					curSumU.second = (number(1) << (index + 1)) - 1;
					u += 2;
				}

				continue;
			}

			if (curPrime.p * curPrime.p >= u_end)
				break;

			number u = u_begin % curPrime.p;
			if (u == 0)
				u = u_begin;
			else
				u = u_begin + curPrime.p - u;

			while (u < u_end)
			{
				std::pair<number, number>& curSumU = sumsU[u - u_begin];
				curSumU.first *= curPrime.inverse;
				number k = u * curPrime.inverse;
				number n = curPrime.p;
				number curSum = curPrime.p + 1;
				for (;;)
				{
					const number q = k * curPrime.inverse;
					if (q > curPrime.max_value)
						break;
					n *= curPrime.p;
					curSum += n;
					k = q;
					curSumU.first *= curPrime.inverse;
				}
				curSumU.second *= curSum;
				u += curPrime.p;
			}
		}

		for (number u = u_begin, u_mod_6 = u_begin % 6; u < u_end; ++u, IncrementMod6(u_mod_6))
		{
			// skip u if it's not coprime to 6
			if ((1 << u_mod_6) & ((1 << 0) | (1 << 2) | (1 << 3) | (1 << 4)))
				continue;

			const std::pair<number, number>& curSumU = sumsU[u - u_begin];
			number sumU = curSumU.second;
			if (curSumU.first > 1)
				sumU *= curSumU.first + 1;

			if ((u - 1) * 2 > sumU)
			{
				// 0.5 < (u - 1) / sumU < 1
				// 1 < (u - 1) * 2 / sumU < 2
				// 0 < (u - 1) * 2 / sumU - 1 < 1
				// 0 < ((u - 1) * 2 - sumU) / sumU < 1
				number r;
				number q = udiv128((u - 1) * 2 - sumU, 0, sumU, &r);
				if (r * 2 > sumU)
					++q;

				const number sumIndex = FindSum(q);
				if (sumIndex < count)
				{
					breeders << q << ' ' << u << ' ' << sumU << std::endl;
					breeders.flush();
				}
			}
		}
	}

	NOINLINE void SearchU(int argc, char* argv[])
	{
		number NumThreads = static_cast<number>((argc < 3) ? 8 : atoi(argv[2]));
		if (NumThreads > 16)
			NumThreads = 16;

		const number step = 10000000;
		
		if (argc > 1)
			URangeData::counter = (number(atoi(argv[1])) * 1000000000) / step;

		std::vector<URangeData> data(NumThreads);
		std::vector<HANDLE> handles(NumThreads);
		std::vector<std::fstream> breeders(NumThreads);
		for (number i = 0; i < NumThreads; ++i)
		{
			URangeData& cur = data[i];
			cur.threadIndex = i;
			cur.numThreads = NumThreads;
			cur.pThis = this;
			cur.step = step;
			cur.breeders = &breeders[i];
			char name[64];
			sprintf_s(name, "breeders%I64u.txt", i);
			breeders[i].open(name, std::ios_base::out);
			handles[i] = (HANDLE) _beginthreadex(0, 0, SearchURangeThreadProc, &cur, 0, 0);
		}
		WaitForMultipleObjects(static_cast<DWORD>(NumThreads), handles.data(), true, INFINITE);
		for (number i = 0; i < NumThreads; ++i)
			CloseHandle(handles[i]);

		UnmapViewOfFile(begin);
		CloseHandle(hSumsMap);
		CloseHandle(hSumsFile);

		printf("\nSearch for U values terminated at %I64u\n", URangeData::counter * step);
		
		std::unordered_map<number, std::vector<std::pair<number, number>>> found_u;
		for (number i = 0; i < NumThreads; ++i)
		{
			char name[64];
			sprintf_s(name, "breeders%I64u.txt", i);
			std::ifstream f(name);
			while (!f.eof())
			{
				number q = 0;
				number u = 0;
				number sumU = 0;
				f >> q >> u >> sumU;
				if (!q || !u || !sumU)
					break;
				found_u[q].push_back(std::pair<number, number>(u, sumU));
			}
		}
		std::ofstream breeders_found("breeders_found.txt");
		count = 0;
		FindBreeders(1, 1, SearchLimit::SQRT2, 0, breeders_found, found_u);
		breeders_found.close();
	}

	NOINLINE static void OutputBreeder(std::ostream& s, number u, number a)
	{
		number sumU;
		std::vector<std::pair<number, number>> fU;
		Factorize(u, sumU, fU);

		number sumA;
		std::vector<std::pair<number, number>> fA;
		Factorize(a, sumA, fA);

		number sumSumU;
		std::vector<std::pair<number, number>> fSU;
		Factorize(sumU, sumSumU, fSU);

		number u1 = sumU - 1 + u;
		number sumU1;
		std::vector<std::pair<number, number>> fU1;
		Factorize(u1, sumU1, fU1);

		std::vector<std::pair<number, number>> rightSideAP = fSU;
		rightSideAP.insert(rightSideAP.end(), fU1.begin(), fU1.end());
		NormalizeFactorization(rightSideAP);

		std::vector<std::pair<number, number>> rightSideBreeder = fU1;
		number u_p1 = u + 1;
		number sum_UP1;
		std::vector<std::pair<number, number>> fU_p1;
		Factorize(u_p1, sum_UP1, fU_p1);
		rightSideBreeder.insert(rightSideBreeder.end(), fU_p1.begin(), fU_p1.end());
		NormalizeFactorization(rightSideBreeder);

		s << u << ' ' << a << ' ' << (sumU - 1) << ' ';
		PrintFactorization(s, fU);
		s << ' ';
		PrintFactorization(s, fA);
		s << ' ';
		PrintFactorization(s, fSU);
		s << ' ';
		PrintFactorization(s, rightSideAP);
		s << ' ';
		PrintFactorization(s, rightSideBreeder);
		s << std::endl;
	}
} locSumsA;

volatile unsigned long long FoundSums::URangeData::counter = 0;
bool FoundSums::StopSearchURanges = false;

int main(int argc, char* argv[])
{
	PrimeTablesInit();

	// No command line arguments: start the search from u = 0
	// One command line argument (k): start the search from u = k*10^9, sums.dat must have been already generated
	locSumsA.Init(argc == 1);
	locSumsA.SearchU(argc, argv);
}
